from .BaseClient import BaseClient
from .OCSClient import OCSClient
from .EDSClient import EDSClient
from .Types import Types
from .Streams import Streams
from .DataViews import DataViews
from .SdsError import SdsError
from .SDS import *
from .DataView import *



class SdsError(Exception):
    """
    Helper class to hold exceptions
    """

    def __init__(self, value):
        """
        Set thee exception value
        :param value:
        """
        self.value = value

    def __str__(self):
        """
        Get the exception
        :return:
        """
        return repr(self.value)



from .SdsBoundaryType import SdsBoundaryType
from .SdsNamespace import SdsNamespace
from .SdsSearchMode import SdsSearchMode
from .SdsStream import SdsStream
from .SdsStreamExtrapolation import SdsStreamExtrapolation
from .SdsStreamIndex import SdsStreamIndex
from .SdsStreamMode import SdsStreamMode
from .SdsStreamPropertyOverride import SdsStreamPropertyOverride
from .SdsStreamView import SdsStreamView
from .SdsStreamViewMap import SdsStreamViewMap
from .SdsStreamViewProperty import SdsStreamViewProperty
from .SdsType import SdsType
from .SdsTypeCode import SdsTypeCode
from .SdsTypeProperty import SdsTypeProperty



from .BaseClient import BaseClient
from .DataViews import DataViews
from .Types import Types
from .Streams import Streams


class OCSClient:
    """
    A client that handles communication with OCS
    """

    def __init__(self, apiversion, tenant, url, clientId, clientSecret=None,
                 acceptVerbosity=False):
        """
        Use this to help communinication with OCS
        :param apiversion: Version of the api you are communicating with
        :param tenant: Your tenant ID
        :param url: The base URL for your OCS instance
        :param clientId: Your client ID
        :param clientSecret: Your client Secret or Key
        :param acceptVerbosity: Sets whether in value calls you get all values
                or just non-default values
        """
        self.__baseClient = BaseClient(apiversion, tenant, url, clientId,
                                       clientSecret, acceptVerbosity)
        self.__DataViews = DataViews(self.__baseClient)
        self.__Types = Types(self.__baseClient)
        self.__Streams = Streams(self.__baseClient)

    @property
    def uri(self):
        """
        :return: The uri of this OCS client as a string
        """
        return self.__baseClient.uri

    @property
    def tenant(self):
        """
        :return: The tenant of this OCS client as a string
        """
        return self.__baseClient.tenant

    @property
    def acceptverbosity(self):
        """
        :return: Whether this will include the accept verbosity header
        """
        return self.__baseClient.AcceptVerbosity

    @acceptverbosity.setter
    def acceptverbosity(self, AcceptVerbosity):
        self.__baseClient.AcceptVerbosity = AcceptVerbosity

    @property
    def request_timeout(self):
        """
        :return: Request timeout in seconds (default 30 secs)
        """
        return self.__baseClient.RequestTimeout

    @request_timeout.setter
    def request_timeout(self, timeout):
        self.__baseClient.RequestTimeout = timeout

    @property
    def DataViews(self):
        """
        :return: A client for interacting with Data Views
        """
        return self.__DataViews

    @property
    def Types(self):
        """
        :return: A client for interacting with Types
        """
        return self.__Types

    @property
    def Streams(self):
        """
        :return: A client for interacting with Streams
        """
        return self.__Streams

    @property
    def baseClient(self):
        """
        :return: A client for interacting with the baseclient directly
        """
        return self.__baseClient



import json
from .DataItem import DataItem


class DataItems(object):

    def __init__(
        self,
        timeOfResolution=None,
        items=None
    ):
        """
        :param timeOfResolution: not required
        :param items: not required
        """
        self.__timeOfResolution = timeOfResolution
        if items:
            self.__items = items
        else:
            self.__items = []

    @property
    def TimeOfResolution(self):
        """
        Get the timeOfResolution  required
        :return:
        """
        return self.__timeOfResolution

    @TimeOfResolution.setter
    def TimeOfResolution(self, timeOfResolution):
        """
        Set the timeOfResolution  required
        :param timeOfResolution:
        :return:
        """
        self.__timeOfResolution = timeOfResolution

    @property
    def Items(self):
        """
        Get the items  required
        :return:
        """
        return self.__items

    @Items.setter
    def Items(self, items):
        """
        Set the items  required
        :param items:
        :return:
        """
        self.__items = items

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {}

        # optional properties
        if hasattr(self, 'TimeOfResolution'):
            dictionary['TimeOfResolution'] = self.TimeOfResolution

        if hasattr(self, "Items"):
            dictionary["Items"] = []
            for value in self.Items:
                dictionary["Items"].append(value.toDictionary())

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return DataItems.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        dataItems = DataItems()

        if not content:
            return dataItems

        if 'TimeOfResolution' in content:
            dataItems.TimeOfResolution = content['TimeOfResolution']

        if "Items" in content:
            Items = content["Items"]
            if Items is not None and len(Items) > 0:
                dataItems.Items = []
                for value in Items:
                    dataItems.Items.append(
                        DataItem.fromDictionary(value))

        return dataItems



import json
from .SdsStreamViewProperty import SdsStreamViewProperty


class SdsStreamViewMap(object):
    """
    SdsStreamViewMap definitions
    """

    @property
    def SourceTypeId(self):
        """
        required
        :return:
        """
        return self.__sourceTypeId

    @SourceTypeId.setter
    def SourceTypeId(self, baseType):
        """
        required
        :param baseType:
        :return:
        """
        self.__sourceTypeId = baseType

    @property
    def TargetTypeId(self):
        """
        required
        :return:
        """
        return self.__targetTypeId

    @TargetTypeId.setter
    def TargetTypeId(self, typeCode):
        """
        required
        :param typeCode:
        :return:
        """
        self.__targetTypeId = typeCode

    @property
    def Properties(self):
        """
        list of SdsStreamViewMapProperty    not required
        :return:
        """
        return self.__properties

    @Properties.setter
    def Properties(self, properties):
        """
        list of SdsStreamViewMapProperty    not required
        :param properties:
        :return:
        """
        self.__properties = properties

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {'SourceTypeId': self.SourceTypeId,
                      'TargetTypeId': self.TargetTypeId}

        # optional properties
        if hasattr(self, 'Properties'):
            dictionary['Properties'] = []
            for value in self.Properties:
                dictionary['Properties'].append(value.toDictionary())

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return SdsStreamViewMap.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        streamViewMap = SdsStreamViewMap()

        if not content:
            return streamViewMap

        if 'TargetTypeId' in content:
            streamViewMap.TargetTypeId = content['TargetTypeId']

        if 'SourceTypeId' in content:
            streamViewMap.SourceTypeId = content['SourceTypeId']

        if 'Properties' in content:
            properties = content['Properties']
            if properties is not None and len(properties) > 0:
                streamViewMap.Properties = []
                for value in properties:
                    streamViewMap.Properties.append(
                        SdsStreamViewProperty.fromDictionary(value))

        return streamViewMap



import json


class Query(object):

    def __init__(
        self,
        id=None,
        value=None
    ):
        """
        :param id: required
        :param value: not required
        """
        self.__id = id
        self.__value = value

    @property
    def Id(self):
        """
        Get the id  required
        :return:
        """
        return self.__id

    @Id.setter
    def Id(self, id):
        """
        Set the id  required
        :param id:
        :return:
        """
        self.__id = id

    @property
    def Value(self):
        """
        Get the value  required
        :return:
        """
        return self.__value

    @Value.setter
    def Value(self, value):
        """
        Set the value  required
        :param value:
        :return:
        """
        self.__value = value

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {'Id': self.Id}

        # optional properties
        if hasattr(self, 'Value'):
            dictionary['Value'] = self.Value

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return Query.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        query = Query()

        if not content:
            return query

        if 'Id' in content:
            query.Id = content['Id']

        if 'Value' in content:
            query.Value = content['Value']

        return query



from enum import Enum


class SdsStreamMode(Enum):
    """
    Enum 0-3
    """
    Continuous = 0
    StepwiseContinuousLeading = 1
    StepwiseContinuousTrailing = 2
    Discrete = 3
    Default = Continuous



import json
from .SdsTypeCode import SdsTypeCode
from .SdsTypeProperty import SdsTypeProperty


class SdsType(object):
    """Sds type definitions"""

    def __init__(self, id=None, name=None, description=None, baseType=None,
                 sdsTypeCode=SdsTypeCode.Empty, properties=[]):
        """

        :param id: required
        :param name: not required
        :param description: not required
        :param baseType: not required
        :param sdsTypeCode: SdsTypeCode    required
        :param properties: array of SdsTypeProperty   required
        """
        self.SdsTypeCode = sdsTypeCode
        self.Id = id
        self.Name = name
        self.Description = description
        self.BaseType = baseType
        self.Properties = properties

    @property
    def Id(self):
        """
        required
        :return:
        """
        return self.__id

    @Id.setter
    def Id(self, id):
        """
        required
        :param id:
        :return:
        """
        self.__id = id

    @property
    def Name(self):
        """
        not required
        :return:
        """
        return self.__name

    @Name.setter
    def Name(self, name):
        """
        not required
        :param name:
        :return:
        """
        self.__name = name

    @property
    def Description(self):
        """
        not required
        :return:
        """
        return self.__description

    @Description.setter
    def Description(self, description):
        """
        not required
        :param description:
        :return:
        """
        self.__description = description

    @property
    def BaseType(self):
        """
        not required
        :return:
        """
        return self.__baseType

    @BaseType.setter
    def BaseType(self, baseType):
        """
        not required
        :param baseType:
        :return:
        """
        self.__baseType = baseType

    @property
    def SdsTypeCode(self):
        """
        required
        :return:
        """
        return self.__typeCode

    @SdsTypeCode.setter
    def SdsTypeCode(self, typeCode):
        """
        SdsTypeCode    required
        :param typeCode:
        :return:
        """
        self.__typeCode = typeCode

    @property
    def Properties(self):
        """
        SdsTypeCode  required
        :return:
        """
        return self.__properties

    @Properties.setter
    def Properties(self, properties):
        """
        required
        :param properties:
        :return:
        """
        self.__properties = properties

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {'SdsTypeCode': self.SdsTypeCode.value}

        # optional properties
        if hasattr(self, 'Properties'):
            if(self.Properties):
                dictionary['Properties'] = []
                for value in self.Properties:
                    dictionary['Properties'].append(value.toDictionary())

        if hasattr(self, 'Id'):
            dictionary['Id'] = self.Id

        if hasattr(self, 'Name'):
            dictionary['Name'] = self.Name

        if hasattr(self, 'Description'):
            dictionary['Description'] = self.Description

        # if self.BaseType is not None and len(self.BaseType) > 0:
        if hasattr(self, 'BaseType'):
            if(self.BaseType is not None):
                dictionary['BaseType'] = self.BaseType.toDictionary()

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return SdsType.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        type = SdsType()

        if not content:
            return type

        if 'Id' in content:
            type.Id = content['Id']

        if 'Name' in content:
            type.Name = content['Name']

        if 'Description' in content:
            type.Description = content['Description']

        if 'SdsTypeCode' in content:
            type.SdsTypeCode = SdsTypeCode(content['SdsTypeCode'])

        if 'BaseType' in content:
            type.BaseType = SdsType.fromDictionary(content['BaseType'])

        if 'Properties' in content:
            properties = content['Properties']
            if properties is not None and len(properties) > 0:
                type.Properties = []
                for value in properties:
                    type.Properties.append(
                        SdsTypeProperty.fromDictionary(value))

        return type



import json
from .SdsStreamIndex import SdsStreamIndex
from .SdsStreamPropertyOverride import SdsStreamPropertyOverride


class SdsStream(object):
    """Sds stream definition"""

    def __init__(self, id=None, name=None, description=None, typeId=None,
                 propertyOverrides=None, indexes=None, interpolationMode=None,
                 extrapolationMode=None):
        """

        :param id: required
        :param name: not required
        :param description: not required
        :param typeId: required
        :param propertyOverrides:  array of  SdsStreamPropertyOverride
                                   not required
        :param indexes: array of SdsStreamIndex   not required
        :param interpolationMode: SdsInterpolationMode   default is null
                                   not required
        :param extrapolationMode: SdsExtrapolationMode default is null
                                  not required
        """
        self.Id = id
        self.Name = name
        self.Description = description
        self.TypeId = typeId
        self.PropertyOverrides = propertyOverrides
        self.Indexes = indexes
        self.InterpolationMode = interpolationMode
        self.ExtrapolationMode = extrapolationMode

    @property
    def Id(self):
        """
        required
        :return:
        """
        return self.__id

    @Id.setter
    def Id(self, id):
        """"
        required
        :param id:
        :return:
        """
        self.__id = id

    @property
    def Name(self):
        """
        not required
        :return:
        """
        return self.__name

    @Name.setter
    def Name(self, name):
        """
        not required
        :param name:
        :return:
        """
        self.__name = name

    @property
    def Description(self):
        """
        not required
        :return:
        """
        return self.__description

    @Description.setter
    def Description(self, description):
        """
        not required
        :param description:
        :return:
        """
        self.__description = description

    @property
    def TypeId(self):
        """
        required
        :return:
        """
        return self.__typeId

    @TypeId.setter
    def TypeId(self, typeId):
        """
        required
        :param typeId:
        :return:
        """
        self.__typeId = typeId

    @property
    def PropertyOverrides(self):
        """
        array of  SdsStreamPropertyOverride    not required
        :return:
        """
        return self.__propertyOverrides

    @PropertyOverrides.setter
    def PropertyOverrides(self, propertyOverrides):
        """
        array of  SdsStreamPropertyOverride    not required
        :param propertyOverrides:
        :return:
        """
        self.__propertyOverrides = propertyOverrides

    @property
    def Indexes(self):
        """
        array of SdsStreamIndex   not required
        :return:
        """
        return self.__indexes

    @Indexes.setter
    def Indexes(self, indexes):
        """
        array of SdsStreamIndex   not required
        :param indexes:
        :return:
        """
        self.__indexes = indexes

    @property
    def InterpolationMode(self):
        """
        SdsInterpolationMode   default is null   not required
        :return:
        """
        return self.__interpolationMode

    @InterpolationMode.setter
    def InterpolationMode(self, interpolationMode):
        """
        SdsInterpolationMode    default is null   not required
        :param interpolationMode:
        :return:
        """
        self.__interpolationMode = interpolationMode

    @property
    def ExtrapolationMode(self):
        """
        SdsExtrapolationMode default is null   not required
        :return:
        """
        return self.__extrapolationMode

    @ExtrapolationMode.setter
    def ExtrapolationMode(self, extrapolationMode):
        """
        SdsExtrapolationMode default is null   not required
        :param extrapolationMode:
        :return:
        """
        self.__extrapolationMode = extrapolationMode

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {'Id': self.Id, 'TypeId': self.TypeId}

        # optional properties
        if hasattr(self, 'Name'):
            dictionary['Name'] = self.Name

        if hasattr(self, 'Description'):
            dictionary['Description'] = self.Description

        if hasattr(self, 'InterpolationMode'):
            dictionary['InterpolationMode'] = self.InterpolationMode

        if hasattr(self, 'ExtrapolationMode'):
            dictionary['ExtrapolationMode'] = self.ExtrapolationMode

        if hasattr(self, 'PropertyOverrides'):
            if self.PropertyOverrides is not None:
                dictionary['PropertyOverrides'] = []
                for value in self.PropertyOverrides:
                    dictionary['PropertyOverrides'].append(
                        value.toDictionary())

        if hasattr(self, 'Indexes'):
            if self.Indexes is not None:
                dictionary['Indexes'] = []
                for value in self.Indexes:
                    dictionary['Indexes'].append(value.toDictionary())

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return SdsStream.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        stream = SdsStream()

        if not content:
            return stream

        if 'Id' in content:
            stream.Id = content['Id']

        if 'Name' in content:
            stream.Name = content['Name']

        if 'Description' in content:
            stream.Description = content['Description']

        if 'InterpolationMode' in content:
            stream.InterpolationMode = content['InterpolationMode']

        if 'ExtrapolationMode' in content:
            stream.ExtrapolationMode = content['ExtrapolationMode']

        if 'TypeId' in content:
            stream.TypeId = content['TypeId']

        if 'PropertyOverrides' in content:
            propertyOverrides = content['PropertyOverrides']
            if propertyOverrides is not None and len(propertyOverrides) > 0:
                stream.PropertyOverrides = []
                for value in propertyOverrides:
                    stream.PropertyOverrides.append(
                        SdsStreamPropertyOverride.fromDictionary(value))

        if 'Indexes' in content:
            indexes = content['Indexes']
            if indexes is not None and len(indexes) > 0:
                stream.Indexes = []
                for value in indexes:
                    stream.Indexes.append(SdsStreamIndex.fromDictionary(value))

        return stream



import json
from .Field import Field


class DataMapping(object):

    def __init__(
        self,
        targetId=None,
        targetFieldKey=None,
        typeCode=None
    ):
        """
        :param targetId: not required
        :param targetFieldKey: not required
        :param typeCode: not required
        """
        self.__targetId = targetId
        self.__targetFieldKey = targetFieldKey
        self.__typeCode = typeCode

    @property
    def TargetId(self):
        """
        Get the targetId  required
        :return:
        """
        return self.__targetId

    @TargetId.setter
    def TargetId(self, targetId):
        """
        Set the __targetId  required
        :param __targetId:
        :return:
        """
        self.__targetId= targetId

    @property
    def TargetFieldKey(self):
        """
        Get the targetFieldKey  required
        :return:
        """
        return self.__targetFieldKey

    @TargetFieldKey.setter
    def TargetFieldKey(self, targetFieldKey):
        """
        Set the targetFieldKey  required
        :param targetFieldKey:
        :return:
        """
        self.__targetFieldKey = targetFieldKey

    @property
    def TypeCode(self):
        """
        Get the typeCode  required
        :return:
        """
        return self.__typeCode

    @TypeCode.setter
    def TypeCode(self, typeCode):
        """
        Set the typeCode  required
        :param typeCode:
        :return:
        """
        self.__typeCode = typeCode

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {}

        # optional properties
        if hasattr(self, 'TargetId'):
            dictionary['TargetId'] = self.TargetId

        if hasattr(self, 'TargetFieldKey'):
            dictionary['TargetFieldKey'] = self.TargetFieldKey

        if hasattr(self, 'TypeCode'):
            dictionary['TypeCode'] = self.TypeCode

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return DataMapping.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        dataMapping = DataMapping()

        if not content:
            return dataMapping

        if 'TargetId' in content:
            dataMapping.TargetId = content['TargetId']            

        if 'TargetFieldKey' in content:
            dataMapping.TargetFieldKey = content['TargetFieldKey']            

        if 'TypeCode' in content:
            dataMapping.TypeCode = content['TypeCode']

        return dataMapping



import json
from .FieldSource import FieldSource


class Field(object):

    def __init__(
        self,
        source=None,
        keys=None,
        label=None,
    ):
        """

        :param source: not required
        :param keys: not required
        :param label: not required
        """
        self.__source = source
        if keys:
            self.__keys = keys
        else:
            self.__keys = []
        self.__label = label

    @property
    def Source(self):
        """
        Get the source  required
        :return:
        """
        return self.__source

    @Source.setter
    def Source(self, source):
        """
        Set the source  required
        :param source:
        :return:
        """
        self.__source = source

    @property
    def Keys(self):
        """
        Get the keys  required
        :return:
        """
        return self.__keys

    @Keys.setter
    def Keys(self, keys):
        """
        Set the keys  required
        :param keys:
        :return:
        """
        self.__keys = keys

    @property
    def Label(self):
        """
        Get the label  required
        :return:
        """
        return self.__label

    @Label.setter
    def Label(self, label):
        """
        Set the label  required
        :param label:
        :return:
        """
        self.__label = label

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {}

        # optional properties
        if hasattr(self, 'Source'):
            if hasattr(self.Source, 'name'):
                dictionary['Source'] = self.Source.name
            else:
                dictionary['Source'] = self.Source

        if hasattr(self, "Keys"):
            dictionary["Keys"] = []
            for value in self.Keys:
                dictionary["Keys"].append(value)

        if hasattr(self, 'Label'):
            dictionary['Label'] = self.Label

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return Field.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        field = Field()

        if not content:
            return field

        if 'Source' in content:
            field.Source = FieldSource[content['Source']]

        if "Keys" in content:
            Keys = content["Keys"]
            if Keys is not None and len(Keys) > 0:
                field.Keys = []
                for value in Keys:
                    field.Keys.append(value)

        if 'Label' in content:
            field.Label = content['Label']

        return field



from enum import Enum


class SdsTypeProperty(object):
    """
    Sds type property definition
    """

    def __init__(self, id=None, name=None, description=None, isKey=False,
                 sdsType=None, value=None, order=None):
        """

        :param id: required
        :param name: not required
        :param description: not required
        :param isKey: bool   Set whether this property is part of the type's
                     index    required
        :param sdsType: SdsType required
        :param value:  object   not required
        :param order: Integer   Determines the order of a complex index type.
                    If isKey is set and this is part of a complex index
                    this is required.   Not required
        """
        self.Id = id
        self.Description = description
        self.IsKey = isKey
        self.SdsType = sdsType
        self.Value = value
        self.Order = order

    @property
    def Id(self):
        """
        required
        :param self:
        :return:
        """
        return self.__id

    @Id.setter
    def Id(self, id):
        """
        required
        :param self:
        :param id:
        :return:
        """
        self.__id = id

    @property
    def Name(self):
        """
        not required
        :param self:
        :return:
        """
        return self.__name

    @Name.setter
    def Name(self, name):
        """
        not required
        :param self:
        :param name:
        :return:
        """
        self.__name = name

    @property
    def Description(self):
        """
        not required
        :param self:
        :return:
        """
        return self.__description

    @Description.setter
    def Description(self, Description):
        """
        not required
        :param self:
        :param Description:
        :return:
        """
        self.__description = Description

    @property
    def IsKey(self):
        """
        bool  Set whether this property is part of the type's index    required
        :param self:
        :return:
        """
        return self.__isKey

    @IsKey.setter
    def IsKey(self, iskey):
        """
        bool  Set whether this property is part of the type's index   required
        :param self:
        :param iskey:
        :return:
        """
        self.__isKey = iskey

    @property
    def SdsType(self):
        """
        required
        :param self:
        :return:
        """
        return self.__sdsType

    @SdsType.setter
    def SdsType(self, sdsType):
        """
        required
        :param self:
        :param sdsType:
        :return:
        """
        self.__sdsType = sdsType

    @property
    def Value(self):
        """
        not required
        :param self:
        :return:
        """
        return self.__value

    @Value.setter
    def Value(self, value):
        """
        not required
        :param self:
        :param value:
        :return:
        """
        self.__value = value

    @property
    def Order(self):
        """
        Integer   Determines the order of a complex index type.
                  If isKey is set and this is part of a complex index
                  this is required.   Not required
        :param self:
        :return:
        """
        return self.__order

    @Order.setter
    def Order(self, order):
        """
        Integer   Determines the order of a complex index type.  If isKey is
                    set and this is part of a complex index  this is required.
                     Not required
        :param self:
        :param order:
        :return:
        """
        self.__order = order

    def toDictionary(self):
        dictionary = {'IsKey': self.IsKey}

        if hasattr(self, 'Id'):
            dictionary['Id'] = self.Id

        if hasattr(self, 'Name'):
            dictionary['Name'] = self.Name

        if hasattr(self, 'Description'):
            dictionary['Description'] = self.Description

        if hasattr(self, 'SdsType'):
            if(self.SdsType):
                from .SdsType import SdsType
                dictionary['SdsType'] = self.SdsType.toDictionary()

        if hasattr(self, 'Value'):
            if (isinstance(self.Value, Enum)):
                dictionary['Value'] = self.Value.name
            else:
                dictionary['Value'] = self.Value

        if hasattr(self, 'Order'):
            dictionary['Order'] = self.Order

        return dictionary

    @staticmethod
    def fromDictionary(content):
        typeProperty = SdsTypeProperty()

        if not content:
            return typeProperty

        if 'Id' in content:
            typeProperty.Id = content['Id']

        if 'IsKey' in content:
            typeProperty.IsKey = content['IsKey']

        if 'Name' in content:
            typeProperty.Name = content['Name']

        if 'Description' in content:
            typeProperty.Description = content['Description']

        if 'SdsType' in content:
            from .SdsType import SdsType
            typeProperty.SdsType = SdsType.fromDictionary(content['SdsType'])

        if 'Value' in content:
            typeProperty.Value = content['Value']

        if 'Order' in content:
            typeProperty.Order = content['Order']

        return typeProperty



from enum import Enum


class SdsTypeCode(Enum):
    """
    enum 0-712  not fully inclusive
    """
    Empty = 0
    Object = 1
    DBNull = 2
    Boolean = 3
    Char = 4
    SByte = 5
    Byte = 6
    Int16 = 7
    UInt16 = 8
    Int32 = 9
    UInt32 = 10
    Int64 = 11
    UInt64 = 12
    Single = 13
    Double = 14
    Decimal = 15
    DateTime = 16
    String = 18
    Guid = 19
    DateTimeOffset = 20
    TimeSpan = 21
    Version = 22
    NullableBoolean = 103
    NullableChar = 104
    NullableSByte = 105
    NullableByte = 106
    NullableInt16 = 107
    NullableUInt16 = 108
    NullableInt32 = 109
    NullableUInt32 = 110
    NullableInt64 = 111
    NullableUInt64 = 112
    NullableSingle = 113
    NullableDouble = 114
    NullableDecimal = 115
    NullableDateTime = 116
    NullableGuid = 119
    NullableDateTimeOffset = 120
    NullableTimeSpan = 121
    BooleanArray = 203
    CharArray = 204
    SByteArray = 205
    ByteArray = 206
    Int16Array = 207
    UInt16Array = 208
    Int32Array = 209
    UInt32Array = 210
    Int64Array = 211
    UInt64Array = 212
    SingleArray = 213
    DoubleArray = 214
    DecimalArray = 215
    DateTimeArray = 216
    StringArray = 218
    GuidArray = 219
    DateTimeOffsetArray = 220
    TimeSpanArray = 221
    VersionArray = 222
    Array = 400
    IList = 401
    IDictionary = 402
    IEnumerable = 403
    SdsType = 501
    SdsTypeProperty = 502
    SByteEnum = 605
    ByteEnum = 606
    Int16Enum = 607
    UInt16Enum = 608
    Int32Enum = 609
    UInt32Enum = 610
    Int64Enum = 611
    UInt64Enum = 612
    NullableSByteEnum = 705
    NullableByteEnum = 706
    NullableInt16Enum = 707
    NullableUInt16Enum = 708
    NullableInt32Enum = 709
    NullableUInt32Enum = 710
    NullableInt64Enum = 711
    NullableUInt64Enum = 712



import json


class SdsNamespace(object):
    """
    definition of SdsNamespace
    """

    @property
    def Id(self):
        return self.__id

    @Id.setter
    def Id(self, id):
        self.__id = id

    def toString(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {'Id': self.Id}

        return dictionary

    @staticmethod
    def fromString(content):
        dictionary = json.loads(content)
        return SdsNamespace.fromDictionary(dictionary)

    @staticmethod
    def fromDictionary(content):
        namespace = SdsNamespace()

        if not content:
            return namespace

        if "Id" in content:
            namespace.Id = content["Id"]

        return namespace



import json


class SdsStreamIndex(object):
    """
    Sds Stream Index definitions
    """

    @property
    def SdsTypePropertyId(self):
        """
        id of property   required
        :return:
        """
        return self.__sdsTypePropertyId

    @SdsTypePropertyId.setter
    def SdsTypePropertyId(self, sdsTypePropertyId):
        """
        id of property   required
        :param sdsTypePropertyId:
        :return:
        """
        self.__sdsTypePropertyId = sdsTypePropertyId

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {'SdsTypePropertyId': self.SdsTypePropertyId}

        return dictionary

    @staticmethod
    def fromDictionary(content):
        typePropertyId = SdsStreamIndex()

        if not content:
            return typePropertyId

        if 'SdsTypePropertyId' in content:
            typePropertyId.SdsTypePropertyId = content['SdsTypePropertyId']

        return typePropertyId




from .SdsError import SdsError

import base64
import configparser
import hashlib
from http.server import HTTPServer, BaseHTTPRequestHandler
import json
import os
import secrets
import time
from urllib.parse import urlparse, parse_qs
import webbrowser

import requests


class Authentication(object):

    def __init__(self, tenant, url, clientId, clientSecret):
        self.__tenant = tenant
        self.__clientId = clientId
        self.__clientSecret = clientSecret
        self.__url = url

        self.__expiration = 0
        self.__token = ""
        if clientSecret is not None:
            self.__getToken = self.__getClientIDSecretToken
        else:
            self.__getToken = self.__getPKCEToken

    def getToken(self):
        if ((self.__expiration - time.time()) > 5 * 60):
            return self.__token

        return self.__getToken()

    def __getClientIDSecretToken(self):
        tokenEndpoint = self.__url + "/identity/connect/token"

        tokenInformation = requests.post(
            tokenEndpoint,
            data={"client_id": self.__clientId,
                  "client_secret": self.__clientSecret,
                  "grant_type": "client_credentials"})

        token = json.loads(tokenInformation.content)

        expiration = token.get("expires_in", None)
        if expiration is None:
            raise SdsError(
                f"Failed to get token, check client id/secret: {token['error']}")

        self.__expiration = float(expiration) + time.time()
        self.__token = token['access_token']
        return self.__token

    def __getPKCEToken(self):
        try:
            redirect_uri = 'http://localhost:5004/callback.html'
            scope = 'openid ocsapi'

            # Set up PKCE Verifier and Code Challenge
            verifier = base64.urlsafe_b64encode(
                secrets.token_bytes(32)).rstrip(b'=')
            challenge = base64.urlsafe_b64encode(
                hashlib.sha256(verifier).digest()).rstrip(b'=')

            # Get OAuth endpoint configuration
            print('Step 1: Get OAuth endpoint configuration...')
            endpoint = json.loads(requests.get(
                self.__url + '/identity/.well-known/openid-configuration').content)
            auth_endpoint = endpoint.get('authorization_endpoint')
            token_endpoint = endpoint.get('token_endpoint')

            # Set up request handler for web browser login
            print('Step 2: Set up server to process authorization response...')

            class RequestHandler(BaseHTTPRequestHandler):
                """Handles authentication redirect uri and extracts authorization code from URL"""
                code = ''

                def do_GET(self):
                    """Handles GET request against this temporary local server"""
                    # Parse out authorization code from query string in request
                    RequestHandler.code = parse_qs(
                        urlparse(self.path).query)['code'][0]

                    # Write response
                    self.send_response(200)
                    self.send_header('Content-Type', 'text/html')
                    self.end_headers()
                    self.wfile.write(
                        '<h1>You can now return to the application.</h1>'.encode())

            # Set up server for web browser login
            server = HTTPServer(('', 5004), RequestHandler)

            # Open web browser against authorization endpoint
            print('Step 3: Authorize the user...')
            auth_url = auth_endpoint + \
                '?response_type=code&code_challenge=' + challenge.decode() + \
                '&code_challenge_method=S256&client_id=' + self.__clientId + \
                '&redirect_uri=' + redirect_uri + \
                '&scope=' + scope + \
                '&acr_values=tenant:' + self.__tenant

            # Open user default web browser at Auth page
            if not webbrowser.open(auth_url):
                raise SdsError(
                    "This notebook/script should be run locally on your machine to authenticate")

            # Wait for response in browser
            print('Step 4: Set server to handle one request...')
            server.handle_request()

            # Use authorization code to get bearer token
            print('Step 5: Get a token using the authorization code...')
            token = requests.post(token_endpoint, [
                ('grant_type', 'authorization_code'),
                ('client_id', self.__clientId),
                ('code_verifier', verifier),
                ('code', RequestHandler.code),
                ('redirect_uri', redirect_uri)])

            token = json.loads(token.content)
            expiration = token.get("expires_in", None)
            if expiration is None:
                raise SdsError(
                    f"Failed to get token, please retry login in: {token['error']}")

            self.__expiration = float(expiration) + time.time()
            self.__token = token['access_token']
            print(f"Step 6: Access token read ok\nComplete!")
            return self.__token

        except Exception as error:
            msg = "Encountered Error: {error}".format(error=error)
            raise SdsError(msg)



import json

from .SDS.SdsType import SdsType


class Types(object):
    """
    Handles communication with Sds Service
    """

    def __init__(self, client):
        self.__apiVersion = client.api_version
        self.__tenant = client.tenant
        self.__url = client.uri_API
        self.__baseClient = client

        self.__setPathAndQueryTemplates()

    def getType(self, namespace_id, type_id):
        """
        Retrieves the type specified by 'type_id' from Sds Service
        :param namespace_id: id of namespace to work against
        :param type_id: id of the type to get
        :return:the type as an SdsType
        """
        if namespace_id is None:
            raise TypeError
        if type_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__typePath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                type_id=type_id))
        self.__baseClient.checkResponse(
            response, f"Failed to get SdsType, {type_id}.")

        _type = SdsType.fromJson(json.loads(response.content))
        response.close()
        return _type

    def getTypeReferenceCount(self, namespace_id, type_id):
        """
        Retrieves the number of times the type is referenced
        :param namespace_id: id of namespace to work against
        :param type_id: id of the type to get references of
        :return: reference count python dynamic object
        """
        if namespace_id is None:
            raise TypeError
        if type_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__typeRefCountPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                type_id=type_id))

        self.__baseClient.checkResponse(
            response, f"Failed to get SdsType reference count, {type_id}.")

        counts = json.loads(response.content)
        response.close()
        return counts

    def getTypes(self, namespace_id, skip=0, count=100, query=""):
        """
        Retrieves a list of types associated with the specified 'namespace_id'
            under the current tenant
        :param namespace_id: id of namespace to work against
        :param skip: number of types to skip, used for paging
        :param count: number of types to retrieve
        :param query: optional query.  Default is ""
        :return: array of types as SdsType
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__typesPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id),
            params={"skip": skip, "count": count, "query": query})
        self.__baseClient.checkResponse(
            response, "Failed to get all SdsTypes.")

        types = json.loads(response.content)
        results = []
        for t in types:
            results.append(SdsType.fromJson(t))
        response.close()
        return results

    def getOrCreateType(self, namespace_id, type):
        """
        Tells Sds Service to create or get a type based on local 'type'
        or get if existing type matches

        :param namespace_id: id of namespace to work against
        :param type:  the SdsType to create or get
        :return:  the created or retrieved SdsType
        """
        if namespace_id is None:
            raise TypeError
        if type is None or not isinstance(type, SdsType):
            raise TypeError
        response = self.__baseClient.request(
            "post",
            self.__typePath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                type_id=type.Id),
            data=type.toJson())
        self.__baseClient.checkResponse(
            response, f"Failed to create type, {type.Id}.")

        type = SdsType.fromJson(json.loads(response.text))
        response.close()
        return type

    def deleteType(self, namespace_id, type_id):
        """
        Tells Sds Service to delete the type specified by 'type_id'

        :param namespace_id: id of namespace to work against
        :param type_id:  id of the type to delete
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if type_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "delete",
            self.__typePath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                type_id=type_id))

        self.__baseClient.checkResponse(
            response, f"Failed to delete SdsType, {type_id}.")

        response.close()

    # private methods

    def __setPathAndQueryTemplates(self):
        """
        used to create needed URI for the other calls
        :return:
        """
        self.__basePath = self.__url + \
            "/Tenants/{tenant_id}/Namespaces/{namespace_id}"
        self.__typesPath = self.__basePath + "/Types"
        self.__typePath = self.__typesPath + "/{type_id}"
        self.__typeRefCountPath = self.__typePath + "/ReferenceCount"



from .DataView import DataView
from .Query import Query
from .DataViewShapes import DataViewShapes
from .Field import Field
from .FieldSet import FieldSet
from .FieldSource import FieldSource 

from .DataItems import DataItems
from .DataItem import DataItem 
from .DataItemField import DataItemField 

from .DataMapping import DataMapping
from .FieldSets import FieldSets
from .FieldMapping import FieldMapping


import json


class DataItemField(object):

    def __init__(
        self,
        id=None,
        name=None,
        typeCode=None,
        isKey=None,
    ):
        """

        """
        self.__id = id
        self.__name = name
        self.__typeCode = typeCode
        self.__isKey = isKey

    @property
    def Id(self):
        """
        Get the id  required
        :return:
        """
        return self.__id

    @Id.setter
    def Id(self, id):
        """
        Set the id  required
        :param id:
        :return:
        """
        self.__id = id

    @property
    def Name(self):
        """
        Get the name  required
        :return:
        """
        return self.__name

    @Name.setter
    def Name(self, name):
        """
        Set the name  required
        :param name:
        :return:
        """
        self.__name = name

    @property
    def TypeCode(self):
        """
        Get the typeCode  required
        :return:
        """
        return self.__typeCode

    @TypeCode.setter
    def TypeCode(self, typeCode):
        """
        Set the typeCode  required
        :param typeCode:
        :return:
        """
        self.__typeCode = typeCode

    @property
    def IsKey(self):
        """
        Get the isKey  required
        :return:
        """
        return self.__isKey

    @IsKey.setter
    def IsKey(self, isKey):
        """
        Set the isKey  required
        :param isKey:
        :return:
        """
        self.__isKey = isKey

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {}

        # optional properties
        if hasattr(self, 'Id'):
            dictionary['Id'] = self.Id

        if hasattr(self, 'Name'):
            dictionary['Name'] = self.Name

        if hasattr(self, 'TypeCode'):
            dictionary['TypeCode'] = self.TypeCode

        if hasattr(self, 'IsKey'):
            dictionary['IsKey'] = self.IsKey

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return DataItemField.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        dataItem = DataItemField()

        if not content:
            return dataItem

        if 'Id' in content:
            dataItem.Id = content['Id']

        if 'Name' in content:
            dataItem.Name = content['Name']

        if 'TypeCode' in content:
            dataItem.TypeCode = content['TypeCode']

        if 'IsKey' in content:
            dataItem.IsKey = content['IsKey']

        return dataItem



import json
import requests
import time

from .Authentication import Authentication
from .SdsError import SdsError


class BaseClient(object):
    """Handles communication with Sds Service.  Internal Use"""

    def __init__(self, apiversion, tenant, url, clientId=None, clientSecret=None,
                 acceptVerbosity=False):
        self.__apiversion = apiversion
        self.__tenant = tenant
        self.__url = url  # if resource.endswith("/")  else resource + "/"
        self.__acceptVerbosity = acceptVerbosity
        self.__requestTimeout = None
        if (clientId is not None):
            self.__auth_object = Authentication(
                tenant, url, clientId, clientSecret)
            self.__auth_object.getToken()
        else:
            self.__auth_object = None

        self.__uri_API = url + '/api/' + apiversion

    @property
    def uri(self):
        """
        Gets the base url
        :return:
        """
        return self.__url

    @property
    def uri_API(self):
        """
        Returns the base URL plus api versioning information
        :return:
        """
        return self.__uri_API

    @property
    def api_version(self):
        """
        Returns just the base api versioning information
        :return:
        """
        return self.__apiversion

    @property
    def tenant(self):
        """
        Returns the tenant ID
        :return:
        """
        return self.__tenant

    @property
    def AcceptVerbosity(self):
        return self.__acceptVerbosity

    @AcceptVerbosity.setter
    def AcceptVerbosity(self, accept_verbosity):
        self.__acceptVerbosity = accept_verbosity

    @property
    def RequestTimeout(self):
        return self.__requestTimeout

    @RequestTimeout.setter
    def RequestTimeout(self, timeout):
        self.__requestTimeout = timeout

    def __getToken(self):
        """
        Gets the bearer token
        :return:
        """
        return self.__auth_object.getToken()

    def sdsHeaders(self):
        """
        Gets the base headers needed for SDS call
        :return:
        """
        headers = {"Content-type": "application/json",
                   "Accept": "application/json"}
        if (self.__auth_object is not None):
            headers['Authorization'] = "Bearer %s" % self.__getToken()
        if (self.__acceptVerbosity):
            headers['Accept-Verbosity'] = "verbose"
        if self.__requestTimeout is not None:
            headers['Request-Timeout'] = str(self.__requestTimeout)

        return headers

    def checkResponse(self, response, main_message):
        if response.status_code < 200 or response.status_code >= 300:
            status = response.status_code
            reason = response.text
            url = response.url

            if "Operation-Id" in response.headers:
                opId = response.headers["Operation-Id"]
                error = f"  {status}:{reason}.  URL {url}  OperationId {opId}"
            else:
                error = f"  {status}:{reason}.  URL {url}"

            response.close()

            message = main_message + error
            raise SdsError(message)

        # this happens on a collection return that is partially successful
        if response.status_code == 207:
            status = response.status_code
            error = response.json["Error"]
            reason = response.json["Reason"]
            errors = str(response.json["ChildErrors"])
            url = response.url

            if "Operation-Id" in response.headers:
                opId = response.headers["Operation-Id"]
                errorToWrite = f"  {status}:{error}:{reason}. \n\n{errors}\n\n  URL {url}  OperationId {opId}"
            else:
                errorToWrite = f"  {status}:{error}:{reason}. \n\n{errors}\n\n  URL {url}"

            response.close()

            message = main_message + errorToWrite
            raise SdsError(message)

    def request(self, method, url, params=None, data=None, headers=None, **kwargs):
        if not headers:
            headers = self.sdsHeaders()
        return requests.request(method, url, params=params, data=data, headers=headers, **kwargs)



import json
import re

from .BaseClient import BaseClient
from .DataView.DataView import DataView
from .DataView.DataItems import DataItems
from .DataView.FieldSets import FieldSets


class DataViews(object):
    """
    Client for interacting with Data Views
    """

    def __init__(self, client: BaseClient):
        """
        Initiliizes the Data View client
        :param client: This is the base client that is used to make the calls
        """
        self.__baseClient = client
        self.__setPathAndQueryTemplates()
        self.__urlLinks = re.compile(r'<(\S+)>; rel="(\S+)"')

    def postDataView(self, namespace_id, dataView):
        """Tells Sds Service to create a Data View based on local 'dataView'
            or get if existing Data View matches
        :param namespace_id: namespace to work against
        :param DataView: Data View definition.  Data View object expected
        :return: Retrieved Data View as Data View object
        """
        if namespace_id is None:
            raise TypeError
        if dataView is None or not isinstance(dataView, DataView):
            raise TypeError

        response = self.__baseClient.request(
            "post",
            self.__dataViewPath.format(
                tenant_id=self.__baseClient.tenant,
                namespace_id=namespace_id,
                dataView_id=dataView.Id,
            ),
            data=dataView.toJson(),
        )

        self.__baseClient.checkResponse(
            response, f"Failed to create Data View, {dataView.Id}."
        )

        dataView = DataView.fromJson(response.json())
        return dataView

    def putDataView(self, namespace_id, dataView):
        """Tells Sds Service to update a Data View based on local 'dataView'
        :param namespace_id: namespace to work against
        :param dataView: Data View definition. Data View object expected
        :return: Retrieved Data View as Data View object
        """
        if namespace_id is None:
            raise TypeError
        if dataView is None or not isinstance(dataView, DataView):
            raise TypeError

        response = self.__baseClient.request(
            "put",
            self.__dataViewPath.format(
                tenant_id=self.__baseClient.tenant,
                namespace_id=namespace_id,
                dataView_id=dataView.Id,
            ),
            data=dataView.toJson(),
        )

        self.__baseClient.checkResponse(
            response, f"Failed to update Data View, {dataView.Id}."
        )

        return

    def deleteDataView(self, namespace_id, dataView_id):
        """
        Tells Sds Service to delete a Data View based on 'dataView_id'
        :param namespace_id: namespace to work against
        :param dataView_id:  id of Data View to delete
        """
        if namespace_id is None:
            raise TypeError
        if dataView_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "delete",
            self.__dataViewPath.format(
                tenant_id=self.__baseClient.tenant,
                namespace_id=namespace_id,
                dataView_id=dataView_id,
            ),
        )

        self.__baseClient.checkResponse(
            response, f"Failed to delete Data View, {dataView_id}."
        )

        return

    def getDataView(self, namespace_id, dataView_id):
        """
        Retrieves the Data View specified by 'dataView_id' from Sds Service
        :param namespace_id: namespace to work against
        :param dataView_id:  id of Data View to get
        :return: Retrieved Data View as Data View object
        """
        if namespace_id is None:
            raise TypeError
        if dataView_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__dataViewPath.format(
                tenant_id=self.__baseClient.tenant,
                namespace_id=namespace_id,
                dataView_id=dataView_id,
            ),
        )

        self.__baseClient.checkResponse(
            response, f"Failed to get Data View, {dataView_id}."
        )

        dataView = DataView.fromJson(response.json())
        return dataView

    def getDataViews(self, namespace_id, skip=0, count=100):
        """
        Retrieves all of the Data Views from Sds Service
        :param namespace_id: namespace to work against
        :param skip: Number of Data Views to skip
        :param count: Number of Data Views to return
        :return: array of Data Views
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__dataViewsPath.format(
                tenant_id=self.__baseClient.tenant, namespace_id=namespace_id
            ),
            params={"skip": skip, "count": count},
        )

        self.__baseClient.checkResponse(response, "Failed to get Data Views.")

        dataViews = json.loads(response.content)

        results = []
        for t in dataViews:
            results.append(DataView.fromJson(t))
        return results

    def getResolvedDataItems(self, namespace_id, dataView_id, query_id):
        """
        Retrieves all of the resolved data items from the specified Data View from
            Sds Service
        :param namespace_id: namespace to work against
        :param dataView_id: Data View to work against
        :param query_id: Query to see data items of
        :return:
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__dataViewResolvedDataItems.format(
                tenant_id=self.__baseClient.tenant,
                namespace_id=namespace_id,
                dataView_id=dataView_id,
                query_id=query_id,
            ),
        )

        self.__baseClient.checkResponse(
            response, f"Failed to get ResolvedDataitems for Data View, {dataView_id}."
        )
        results = DataItems.fromJson(response.json())

        return results

    def getResolvedIneligibleDataItems(self, namespace_id, dataView_id, query_id):
        """
        Retrieves all of the resolved ineligible data items from the specified Data View from
            Sds Service
        :param namespace_id: namespace to work against
        :param dataView_id: Data View to work against
        :param query_id: Query to see data items of
        :return:
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__dataViewResolvedIneligibleDataItems.format(
                tenant_id=self.__baseClient.tenant,
                namespace_id=namespace_id,
                dataView_id=dataView_id,
                query_id=query_id,
            ),
        )

        self.__baseClient.checkResponse(
            response,
            f"Failed to get ResolvedIneligibleDataitems for Data View, {dataView_id}.",
        )
        results = DataItems.fromJson(response.json())

        return results

    def getResolvedAvailableFieldSets(self, namespace_id, dataView_id):
        """
        Retrieves all of the available field sets from the specified Data View from
            Sds Service
        :param namespace_id: namespace to work against
        :param dataView_id: Data View to work against
        :return:
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__dataViewResolvedAvailableFieldSets.format(
                tenant_id=self.__baseClient.tenant,
                namespace_id=namespace_id,
                dataView_id=dataView_id,
            ),
        )

        self.__baseClient.checkResponse(
            response,
            f"Failed to get ResolvedAvailableFieldSetsfor Data View, {dataView_id}.",
        )
        results = FieldSets.fromJson(response.json())

        return results

    def getDataInterpolated(
        self,
        namespace_id=None,
        dataView_id=None,
        count=None,
        form=None,
        startIndex=None,
        endIndex=None,
        interval=None,
        value_class=None,
        url=None,
    ):
        """
        Retrieves the interpolated data of the 'dataView_id' from Sds Service
        :param namespace_id: namespace to work against
        :param dataView_id: Data View to work against
        :param skip: number of values to skip
        :param count: number of values to return
        :param form: form definition
        :param startIndex: start index
        :param endIndex: end index
        :param interval: space between values
        :param value_class: Use this to auto format the data into the defined
            type.  The type is expected to have a fromJson method that takes a
            dynamicObject and converts it into the defined type.
            Otherwise you get a dynamic object
        :return:
        """
        if url is None:
            if namespace_id is None:
                raise TypeError
            if dataView_id is None:
                raise TypeError

        params = {
            "count": count,
            "form": form,
            "startIndex": startIndex,
            "endIndex": endIndex,
            "interval": interval,
        }
        response = {}
        if url:
            response = self.__baseClient.request("get", url)
        else:
            response = self.__baseClient.request(
                "get",
                self.__dataViewDataInterpolated.format(
                    tenant_id=self.__baseClient.tenant,
                    namespace_id=namespace_id,
                    dataView_id=dataView_id,
                ),
                params=params,
            )

        self.__baseClient.checkResponse(
            response,
            f"Failed to get Data View data interpolated for Data View, {dataView_id}.",
        )

        # build dictionary of first/next page URL links, if any
        links_header = response.headers.get("Link", "")
        links = {
            link.group(2): link.group(1)
            for link in self.__urlLinks.finditer(links_header)
        }

        nextPage = links.get("next", None)
        firstPage = links.get("first", None)

        if form is not None:
            return response.text, nextPage, firstPage

        content = response.json()

        if value_class is None:
            return content, nextPage, firstPage
        return value_class.fromJson(content), nextPage, firstPage

    def getDataStored(
        self,
        namespace_id=None,
        dataView_id=None,
        startIndex=None,
        endIndex=None,
        value_class=None,
        count=None,
        form=None,
        url=None,
    ):
        """
        Retrieves the stored data of the 'dataView_id' from Sds Service
        :param namespace_id: namespace to work against
        :param dataView_id: Data View to work against
        :param count: number of values to return
        :param form: form definition (only 'tableh' is available)
        :param startIndex: start index
        :param endIndex: end index
        :param value_class: Use this to auto format the data into the defined
            type.  The type is expected to have a fromJson method that takes a
            dynamicObject and converts it into the defined type.
            Otherwise you get a dynamic object
        :return:
        """
        if url is None:
            if namespace_id is None:
                raise TypeError
            if dataView_id is None:
                raise TypeError

        params = {
            "count": count,
            "form": form,
            "startIndex": startIndex,
            "endIndex": endIndex,
        }
        response = {}
        if url:
            response = self.__baseClient.request("get", url)
        else:
            response = self.__baseClient.request(
                "get",
                self.__dataViewDataStored.format(
                    tenant_id=self.__baseClient.tenant,
                    namespace_id=namespace_id,
                    dataView_id=dataView_id,
                ),
                params=params,
            )

        self.__baseClient.checkResponse(
            response,
            f"Failed to get Data View data stored for Data View, {dataView_id}.",
        )

        # build dictionary of first/next page URL links, if any
        links_header = response.headers.get("Link", "")
        links = {
            link.group(2): link.group(1)
            for link in self.__urlLinks.finditer(links_header)
        }

        nextPage = links.get("next", None)
        firstPage = links.get("first", None)

        if form is not None:
            return response.text, nextPage, firstPage

        content = response.json()

        if value_class is None:
            return content, nextPage, firstPage
        return value_class.fromJson(content), nextPage, firstPage

    def __setPathAndQueryTemplates(self):
        """
        Internal  Sets the needed URLs
        :return:
        """
        self.__basePath = (
            self.__baseClient.uri_API + "/Tenants/{tenant_id}/Namespaces/{namespace_id}"
        )

        self.__dataViewsPath = self.__basePath + "/dataviews"
        self.__dataViewPath = self.__dataViewsPath + "/{dataView_id}"
        self.__dataViewResolved = self.__dataViewPath + "/Resolved"
        self.__dataViewResolvedDataItems = (
            self.__dataViewResolved + "/DataItems/{query_id}"
        )
        self.__dataViewResolvedIneligibleDataItems = (
            self.__dataViewResolved + "/IneligibleDataItems/{query_id}"
        )
        self.__dataViewResolvedAvailableFieldSets = (
            self.__dataViewResolved + "/AvailableFieldSets"
        )
        self.__dataViewData = self.__dataViewPath + "/data"
        self.__dataViewDataInterpolated = self.__dataViewData + "/interpolated"
        self.__dataViewDataStored = self.__dataViewData + "/stored"



import json
from .DataItemField import DataItemField


class DataItem(object):

    def __init__(
        self,
        resourceType=None,
        id=None,
        name=None,
        typeId=None,
        tags=None,
        metadata=None,
        dataItemFields=None
    ):
        """

        """
        self.__resourceType = resourceType
        self.__id = id
        self.__name = name
        self.__typeId = typeId

        if tags:
            self.__tags = tags
        else:
            self.__tags = []

        self.__metadata = metadata

        if dataItemFields:
            self.__dataItemFields = dataItemFields
        else:
            self.__dataItemFields = []

    @property
    def ResourceType(self):
        """
        Get the resourceType  required
        :return:
        """
        return self.__resourceType

    @ResourceType.setter
    def ResourceType(self, resourceType):
        """
        Set the resourceType  required
        :param resourceType:
        :return:
        """
        self.__resourceType = resourceType

    @property
    def Id(self):
        """
        Get the id  required
        :return:
        """
        return self.__id

    @Id.setter
    def Id(self, id):
        """
        Set the id  required
        :param id:
        :return:
        """
        self.__id = id

    @property
    def Name(self):
        """
        Get the name  required
        :return:
        """
        return self.__name

    @Name.setter
    def Name(self, name):
        """
        Set the name  required
        :param name:
        :return:
        """
        self.__name = name

    @property
    def TypeId(self):
        """
        Get the typeId  required
        :return:
        """
        return self.__typeId

    @TypeId.setter
    def TypeId(self, typeId):
        """
        Set the typeId  required
        :param typeId:
        :return:
        """
        self.__typeId = typeId

    @property
    def Tags(self):
        """
        Get the tags  required
        :return:
        """
        return self.__tags

    @Tags.setter
    def Tags(self, tags):
        """
        Set the tags  required
        :param tags:
        :return:
        """
        self.__tags = tags

    @property
    def Metadata(self):
        """
        Get the metadata  required
        :return:
        """
        return self.__metadata

    @Metadata.setter
    def Metadata(self, metadata):
        """
        Set the metadata  required
        :param metadata:
        :return:
        """
        self.__metadata = metadata

    @property
    def DataItemFields(self):
        """
        Get the dataItemFields  required
        :return:
        """
        return self.__dataItemFields

    @DataItemFields.setter
    def DataItemFields(self, dataItemFields):
        """
        Set the dataItemFields  required
        :param dataItemFields:
        :return:
        """
        self.__dataItemFields = dataItemFields

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {}

        # optional properties
        if hasattr(self, 'ResourceType'):
            dictionary['ResourceType'] = self.ResourceType

        if hasattr(self, 'Id'):
            dictionary['Id'] = self.Id

        if hasattr(self, 'Name'):
            dictionary['Name'] = self.Name

        if hasattr(self, 'TypeId'):
            dictionary['TypeId'] = self.TypeId

        if hasattr(self, "Tags"):
            dictionary["Tags"] = []
            for value in self.Tags:
                dictionary["Tags"].append(value)

        if hasattr(self, 'Metadata'):
            # this is an object of some form we need to deal with better #fixit
            dictionary['Metadata'] = self.Metadata

        if hasattr(self, "DataItemFields"):
            dictionary["DataItemFields"] = []
            for value in self.DataItemFields:
                dictionary["DataItemFields"].append(value.toDictionary())

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return DataItem.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        dataItem = DataItem()

        if not content:
            return dataItem

        if 'ResourceType' in content:
            dataItem.ResourceType = content['ResourceType']

        if 'Id' in content:
            dataItem.Id = content['Id']

        if 'Name' in content:
            dataItem.Name = content['Name']

        if 'TypeId' in content:
            dataItem.TypeId = content['TypeId']

        if "Tags" in content:
            Tags = content["Tags"]
            if Tags is not None and len(Tags) > 0:
                dataItem.Tags = []
                for value in Tags:
                    dataItem.Tags.append(value)

        if 'Metadata' in content:
            # this is an object of some form we need to deal with better #fixit
            dataItem.Metadata = content['Metadata']

        if "DataItemFields" in content:
            DataItemFields = content["DataItemFields"]
            if DataItemFields is not None and len(DataItemFields) > 0:
                dataItem.DataItemFields = []
                for value in DataItemFields:
                    dataItem.DataItemFields.append(
                        DataItemField.fromDictionary(value))

        return dataItem



import json
from .FieldKind  import FieldKind
from .DataMapping import DataMapping


class FieldMapping(object):

    def __init__(
        self,
        id=None,
        label=None,
        fieldKind=None,
        fieldSetIndex=None,
        fieldIndex=None,
        dataMappings=None
    ):
        """

        :param id: not required
        :param label: not required
        :param fieldKind: not required
        :param fieldSetIndex: not required
        :param fieldIndex: not required
        :param dataMappings: not required
        """
        self.__id = id
        self.__label= label
        if fieldKind:
            self.__fieldKind = fieldKind
        else:
            self.__fieldKind = FieldKind.IndexField
        self.__fieldSetIndex = fieldSetIndex
        self.__fieldIndex = fieldIndex
        self.__dataMappings = dataMappings

    @property
    def Id(self):
        """
        Get the id  required
        :return:
        """
        return self.__id

    @Id.setter
    def Id(self, id):
        """
        Set the id  required
        :param id:
        :return:
        """
        self.__id = id

    @property
    def Label(self):
        """
        Get the label  required
        :return:
        """
        return self.__label

    @Label.setter
    def Label(self, label):
        """
        Set the label  required
        :param label:
        :return:
        """
        self.__label = label

    @property
    def FieldKind(self):
        """
        Get the fieldKind  required
        :return:
        """
        return self.__fieldKind

    @FieldKind.setter
    def FieldKind(self, fieldKind):
        """
        Set the fieldKind  required
        :param fieldKind:
        :return:
        """
        self.__fieldKind = fieldKind

    @property
    def FieldSetIndex(self):
        """
        Get the fieldSetIndex  required
        :return:
        """
        return self.__fieldSetIndex

    @FieldSetIndex.setter
    def FieldSetIndex(self, fieldSetIndex):
        """
        Set the fieldSetIndex  required
        :param fieldSetIndex:
        :return:
        """
        self.__fieldSetIndex = fieldSetIndex

    @property
    def FieldIndex(self):
        """
        Get the fieldIndex  required
        :return:
        """
        return self.__fieldIndex

    @FieldIndex.setter
    def FieldIndex(self, fieldIndex):
        """
        Set the fieldIndex  required
        :param fieldIndex:
        :return:
        """
        self.__fieldIndex = fieldIndex

    @property
    def DataMappings(self):
        """
        Get the dataMappings  required
        :return:
        """
        return self.__dataMappings

    @DataMappings.setter
    def DataMappings(self, dataMappings):
        """
        Set the dataMappings  required
        :param dataMappings:
        :return:
        """
        self.__dataMappings = dataMappings

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {}

        # optional properties
        if hasattr(self, 'Id'):
            dictionary['Id'] = self.Id

        if hasattr(self, 'Label'):
            dictionary['Label'] = self.Label

        if hasattr(self, 'FieldKind'):
            dictionary['FieldKind'] = self.FieldKind.name

        if hasattr(self, 'FieldSetIndex'):
            dictionary['FieldSetIndex'] = self.FieldSetIndex

        if hasattr(self, 'FieldIndex'):
            dictionary['FieldIndex'] = self.FieldIndex

        if hasattr(self, "DataMappings"):
            dictionary["DataMappings"] = []
            for value in self.DataMappings:
                dictionary["DataMappings"].append(value.toDictionary())

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return FieldMapping.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        fieldMapping = FieldMapping()

        if not content:
            return fieldMapping

        if 'Id' in content:
            fieldMapping.Id = content['Id']

        if 'Label' in content:
            fieldMapping.Label = content['Label']

        if 'FieldKind' in content:
            fieldMapping.FieldKind = FieldKind[content['FieldKind']] 

        if 'FieldSetIndex' in content:
            fieldMapping.FieldSetIndex = content['FieldSetIndex']

        if 'FieldIndex' in content:
            fieldMapping.FieldIndex = content['FieldIndex']

        if "DataMapping" in content:
            DataMapping_ = content["DataMapping"]
            if DataMapping_ is not None and len(DataMapping_) > 0:
                fieldMapping.DataFields = []
                for value in DataMapping_:
                    fieldMapping.DataFields.append(
                        DataMapping.fromDictionary(value))

        return fieldMapping



import json
from ..SDS.SdsTypeCode import SdsTypeCode
from .Query import Query
from .FieldSet import FieldSet
from .Field import Field
from .DataViewShapes import DataViewShapes


class DataView(object):
    """
    Data View definition
    """

    def __init__(
        self,
        id=None,
        name=None,
        description=None,
        queries=None,
        indexField=None,
        dataFieldSets=None,
        groupingFields=None,
        indexTypeCode=None,
        defaultStartIndex=None,
        defaultEndIndex=None,
        defaultInterval=None,
        shape=None
    ):
        """
        :param id: required
        :param name: not required
        :param description:  not required
        """
        self.__id = id
        self.__name = name
        self.__description = description
        if queries:
            self.__queries = queries
        else:
            self.__queries = []
        self.__indexField = indexField
        if dataFieldSets:
            self.__dataFieldSets = dataFieldSets
        else:
            self.__dataFieldSets = []
        if groupingFields:
            self.__groupingFields = groupingFields
        else:
            self.__groupingFields = []
        if indexTypeCode:
            self.__indexTypeCode = indexTypeCode
        else:
            self.indexTypeCode = SdsTypeCode.DateTime
        self.__defaultStartIndex = defaultStartIndex
        self.__defaultEndIndex = defaultEndIndex
        self.__defaultInterval = defaultInterval
        if shape:
            self.__shape = shape
        else:
            self.shape = DataViewShapes.Standard

    @property
    def Id(self):
        """
        Get the id  required
        :return:
        """
        return self.__id

    @Id.setter
    def Id(self, id):
        """
        Set the id  required
        :param id:
        :return:
        """
        self.__id = id

    @property
    def Name(self):
        """
        Name can be duplicated in a namespace   not required
        :return:
        """
        return self.__name

    @Name.setter
    def Name(self, name):
        """
        Name can be duplicated in a namespace   not required
        :param name:
        :return:
        """
        self.__name = name

    @property
    def Description(self):
        """
        Add an easy to understand description not required
        :return:
        """
        return self.__description

    @Description.setter
    def Description(self, description):
        """
        Add an easy to understand description not required
        :param description:
        :return:
        """
        self.__description = description

    @property
    def Queries(self):
        """
        Query string  required
        :return:
        """
        return self.__queries

    @Queries.setter
    def Queries(self, queries):
        """
        Array of queries  required
        :param queries:
        :return:
        """
        self.__queries = queries

    @property
    def IndexField (self):
        """
        IndexField field  required
        :return:
        """
        return self.__indexField

    @IndexField.setter
    def IndexField(self, indexField):
        """
        Field indexField  required
        :param indexField:
        :return:
        """
        self.__indexField = indexField

    @property
    def DataFieldSets(self):
        """
        :return:
        """
        return self.__dataFieldSets

    @DataFieldSets.setter
    def DataFieldSets(self, dataFieldSets):
        """
        :param dataFieldSets:
        :return:
        """
        self.__dataFieldSets = dataFieldSets

    @property
    def GroupingFields(self):
        """
        :return:
        """
        return self.__groupingFields

    @GroupingFields.setter
    def GroupingFields(self, groupingFields):
        """
        Array of fields   not required
        :param groupingFields:
        :return:
        """
        self.__groupingFields = groupingFields

    @property
    def IndexTypeCode(self):
        """
        :return:
        """
        return self.__indexTypeCode

    @IndexTypeCode.setter
    def IndexTypeCode(self, indexTypeCode):
        """
        Currently limited to "DateTime" string  required
        :param indexTypeCode:
        :return:
        """
        self.__indexTypeCode = indexTypeCode

    @property
    def DefaultStartIndex(self):
        """
        DefaultStartIndex   not required
        :return:
        """
        return self.__defaultStartIndex

    @DefaultStartIndex.setter
    def DefaultStartIndex(self, defaultStartIndex):
        """DefaultStartIndex
        DefaultStartIndex string  not required
        :param defaultStartIndex:
        :return:
        """
        self.__defaultStartIndex = defaultStartIndex

    @property
    def DefaultEndIndex(self):
        """
        DefaultEndIndex string  not required
        :return:
        """
        return self.__defaultEndIndex

    @DefaultEndIndex.setter
    def DefaultEndIndex(self, defaultEndIndex):
        """DefaultEndIndex
        DefaultEndIndex  string not required
        :param defaultEndIndex:
        :return:
        """
        self.__defaultEndIndex = defaultEndIndex

    @property
    def DefaultInterval(self):
        """
        DefaultInterval string not required
        :return:
        """
        return self.__defaultInterval

    @DefaultInterval.setter
    def DefaultInterval(self, defaultInterval):
        """DefaultStartIndex
        DefaultInterval string not required
        :param defaultInterval:
        :return:
        """
        self.__defaultInterval = defaultInterval

    @property
    def Shape(self):
        """
        Shape string  not required
        :return:
        """
        return self.__shape

    @Shape.setter
    def Shape(self, shape):
        """
        Shape string  not required
        :param shape:
        :return:
        """
        self.__shape = shape

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {"Id": self.Id}

        # optional properties
        if hasattr(self, "Name"):
            dictionary["Name"] = self.Name

        if hasattr(self, "Description"):
            dictionary["Description"] = self.Description

        if hasattr(self, "Queries"):
            dictionary["Queries"] = []
            for value in self.Queries:
                dictionary["Queries"].append(value.toDictionary())

        if hasattr(self, "DataFieldSets"):
            dictionary["DataFieldSets"] = []
            for value in self.DataFieldSets:
                dictionary["DataFieldSets"].append(value.toDictionary())

        if hasattr(self, "GroupingFields"):
            dictionary["GroupingFields"] = []
            for value in self.GroupingFields:
                dictionary["GroupingFields"].append(value.toDictionary())

        if hasattr(self, "IndexTypeCode"):
            dictionary["IndexTypeCode"] = self.IndexTypeCode.name

        if hasattr(self, "DefaultStartIndex"):
            dictionary["DefaultStartIndex"] = self.DefaultStartIndex

        if hasattr(self, "DefaultEndIndex"):
            dictionary["DefaultEndIndex"] = self.DefaultEndIndex

        if hasattr(self, "DefaultInterval"):
            dictionary["DefaultInterval"] = self.DefaultInterval

        if hasattr(self, "Shape"):
            dictionary["Shape"] = self.Shape.name

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return DataView.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        dataView = DataView()

        if not content:
            return dataView

        if "Id" in content:
            dataView.Id = content["Id"]

        if "Name" in content:
            dataView.Name = content["Name"]

        if "Description" in content:
            dataView.Description = content["Description"]

        if "Queries" in content:
            Queries = content["Queries"]
            if Queries is not None and len(Queries) > 0:
                dataView.Queries = []
                for value in Queries:
                    dataView.Queries.append(
                        Query.fromDictionary(value))

        if "DataFieldSets" in content:
            DataFieldSets = content["DataFieldSets"]
            if DataFieldSets is not None and len(DataFieldSets) > 0:
                dataView.DataFieldSets = []
                for value in DataFieldSets:
                    dataView.DataFieldSets.append(
                        FieldSet.fromDictionary(value))

        if "GroupingFields" in content:
            GroupingFields = content["GroupingFields"]
            if GroupingFields is not None and len(GroupingFields) > 0:
                dataView.GroupingFields = []
                for value in GroupingFields:
                    dataView.GroupingFields.append(
                        Field.fromDictionary(value))

        if "IndexTypeCode" in content:
            dataView.IndexTypeCode = SdsTypeCode[content['IndexTypeCode']]

        if "DefaultStartIndex" in content:
            dataView.DefaultStartIndex = content["DefaultStartIndex"]

        if "DefaultEndIndex" in content:
            dataView.DefaultEndIndex = content["DefaultEndIndex"]

        if "DefaultInterval" in content:
            dataView.DefaultInterval = content["DefaultInterval"]

        if "Shape" in content:
            dataView.Shape = DataViewShapes[content['Shape']]

        return dataView



import json.decoder


class SdsStreamViewProperty(object):
    """Sds StreamView Property definition"""
    @property
    def SourceId(self):
        """
        required
        :return:
        """
        return self.__sourceId

    @SourceId.setter
    def SourceId(self, id):
        """
        required
        :param id:
        :return:
        """
        self.__sourceId = id

    @property
    def TargetId(self):
        """
        required
        :return:
        """
        return self.__targetId

    @TargetId.setter
    def TargetId(self, name):
        """
        required
        :param name:
        :return:
        """
        self.__targetId = name

    @property
    def SdsStreamView(self):
        """
        SdsStreamView   not required
        :return:
        """
        return self.__sdsStreamView

    @SdsStreamView.setter
    def SdsStreamView(self, description):
        """
        SdsStreamView   not required
        :param description:
        :return:
        """
        self.__sdsStreamView = description

    @property
    def Mode(self):
        return self.__mode

    @Mode.setter
    def Mode(self, mode):
        self.__mode = mode

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {'SourceId': self.SourceId}

        if hasattr(self, 'TargetId'):
            dictionary['TargetId'] = self.TargetId

        if hasattr(self, 'Mode'):
            dictionary['Mode'] = self.Mode

        if hasattr(self, 'SdsStreamView'):
            from .SdsStreamView import SdsStreamView
            dictionary['SdsStreamView'] = self.SdsStreamView.toDictionary()

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return SdsStreamViewProperty.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        streamViewProperty = SdsStreamViewProperty()

        if not content:
            return streamViewProperty

        if 'SourceId' in content:
            streamViewProperty.SourceId = content['SourceId']

        if 'TargetId' in content:
            streamViewProperty.TargetId = content['TargetId']

        if 'Mode' in content:
            streamViewProperty.Mode = content['Mode']

        if 'SdsStreamView' in content:
            from .SdsStreamView import SdsStreamView
            streamViewProperty.SdsStreamView = SdsStreamView.fromDictionary(
                content['SdsStreamView'])

        return streamViewProperty



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ocs_sample_library_hub",
    version="0.1.19",
    author="OSIsoft",
    license="Apache 2.0",
    author_email="cfoisy@osisoft.com",
    description="OCS (OSIsoft Cloud Services) client library - patched 0.1.17rc0 api/v1 + /stored",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cfoisy-osisoft/sample-ocs-sample_libraries-python",
    packages=setuptools.find_packages(),
    install_requires=[
        'requests>=2.25.1',
        'python-dateutil>=2.8.1'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import json
from .Field import Field


class FieldSet(object):

    def __init__(
        self,
        queryId=None,
        dataFields=None,
        identifyingField=None
    ):
        """
        :param queryId: not required
        :param dataFields: not required
        :param identifyingField: not required
        """
        self.__queryId = queryId
        if dataFields:
            self.__dataFields = dataFields
        else:
            self.__dataFields = []
        self.__identifyingField = identifyingField

    @property
    def QueryId(self):
        """
        Get the queryId  required
        :return:
        """
        return self.__queryId

    @QueryId.setter
    def QueryId(self, queryId):
        """
        Set the queryId  required
        :param queryId:
        :return:
        """
        self.__queryId = queryId

    @property
    def DataFields(self):
        """
        Get the dataFields  required
        :return:
        """
        return self.__dataFields

    @DataFields.setter
    def DataFields(self, dataFields):
        """
        Set the dataFields  required
        :param dataFields:
        :return:
        """
        self.__dataFields = dataFields

    @property
    def IdentifyingField(self):
        """
        Get the identifyingField  required
        :return:
        """
        return self.__identifyingField

    @IdentifyingField.setter
    def IdentifyingField(self, identifyingField):
        """
        Set the identifyingField  required
        :param identifyingField:
        :return:
        """
        self.__identifyingField = identifyingField

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {}

        # optional properties
        if hasattr(self, 'QueryId'):
            dictionary['QueryId'] = self.QueryId

        if hasattr(self, "DataFields"):
            dictionary["DataFields"] = []
            for value in self.DataFields:
                dictionary["DataFields"].append(value.toDictionary())

        if hasattr(self, 'IdentifyingField'):
            if self.IdentifyingField is not None:
                dictionary['IdentifyingField'] = self.IdentifyingField.toDictionary()

        return dictionary

    @staticmethod
    def fromJson(jsonObj):
        return FieldSet.fromDictionary(jsonObj)

    @staticmethod
    def fromDictionary(content):
        fieldSet = FieldSet()

        if not content:
            return fieldSet

        if 'QueryId' in content:
            fieldSet.QueryId = content['QueryId']

        if "DataFields" in content:
            DataFields = content["DataFields"]
            if DataFields is not None and len(DataFields) > 0:
                fieldSet.DataFields = []
                for value in DataFields:
                    fieldSet.DataFields.append(
                        Field.fromDictionary(value))

        if 'IdentifyingField' in content:
            fieldSet.IdentifyingField = Field.fromDictionary(
                content['IdentifyingField'])

        return fieldSet



import json


class SdsStreamPropertyOverride(object):
    """
    Sds Stream PropertyOverride definitions
    """

    @property
    def SdsTypePropertyId(self):
        """
        id   required
        :return:
        """
        return self.__sdsTypePropertyId

    @SdsTypePropertyId.setter
    def SdsTypePropertyId(self, sdsTypePropertyId):
        """
        id   required
        :param sdsTypePropertyId:
        :return:
        """
        self.__sdsTypePropertyId = sdsTypePropertyId

    @property
    def InterpolationMode(self):
        """
        SdsInterpolationMode   not required
        :return:
        """
        return self.__interpolationMode

    @InterpolationMode.setter
    def InterpolationMode(self, interpolationMode):
        """
        SdsInterpolationMode   not required
        :param interpolationMode:
        :return:
        """
        self.__interpolationMode = interpolationMode

    @property
    def Uom(self):
        """
        Unit of Measure    not required
        :return:
        """
        return self.__uom

    @Uom.setter
    def Uom(self, uom):
        """
        Unit of Measure    not required
        :param uom:
        :return:
        """
        self.__uom = uom

    def toJson(self):
        return json.dumps(self.toDictionary())

    def toDictionary(self):
        # required properties
        dictionary = {'SdsTypePropertyId': self.SdsTypePropertyId}

        # optional properties
        if hasattr(self, 'Uom'):
            dictionary['Uom'] = self.Uom

        if hasattr(self, 'InterpolationMode'):
            dictionary['InterpolationMode'] = self.InterpolationMode

        return dictionary

    @staticmethod
    def fromDictionary(content):
        propertyOverride = SdsStreamPropertyOverride()

        if not content:
            return propertyOverride

        if 'SdsTypePropertyId' in content:
            propertyOverride.SdsTypePropertyId = content['SdsTypePropertyId']

        if 'Uom' in content:
            propertyOverride.Uom = content['Uom']

        if 'InterpolationMode' in content:
            propertyOverride.InterpolationMode = content['InterpolationMode']

        return propertyOverride



from enum import Enum


class FieldSource(Enum):
    """
    enum 0-6  fully inclusive
    """
    NotApplicable = 0
    Id = 1
    Name = 2
    PropertyId = 3
    PropertyName = 4
    Tags = 5
    Metadata = 6



import json

from .BaseClient import BaseClient
from .SDS.SdsStream import SdsStream
from .SDS.SdsType import SdsType
from .SDS.SdsStreamView import SdsStreamView
from .SDS.SdsStreamViewMap import SdsStreamViewMap
from .SDS.SdsBoundaryType import SdsBoundaryType


class Streams(object):
    """
    Client for interacting with Streams
    """

    def __init__(self, client: BaseClient):
        """
        :param client: base client that handles auth and base routing
        """
        self.__apiVersion = client.api_version
        self.__tenant = client.tenant
        self.__uri_API = client.uri_API
        self.__baseClient = client

        self.__setPathAndQueryTemplates()

    def getStreamView(self, namespace_id, streamView_id):
        """
        Retrieves the streamView specified by 'streamView_id' from Sds Service
        :param namespace_id: namespace to work against
        :param streamView_id: streamview id to get
        :return: Streamview as SDsStreamview
        """
        if namespace_id is None:
            raise TypeError
        if streamView_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__streamViewsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                streamView_id=streamView_id))

        self.__baseClient.checkResponse(
            response, f"Failed to get SdsStreamView, {streamView_id}.")

        streamView = SdsStreamView.fromJson(json.loads(response.content))
        response.close()
        return streamView

    def getStreamViewMap(self, namespace_id, streamView_id):
        """
        Retrieves the streamView map specified by 'streamView_id' from Sds
            Service
        :param namespace_id: namespace to work against
        :param streamView_id:  streamview map to get
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if streamView_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__streamViewsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                streamView_id=streamView_id) + "/Map")

        self.__baseClient.checkResponse(
            response, f"Failed to get SdsStreamViewMap, {streamView_id}.")

        streamViewMap = SdsStreamViewMap.fromJson(json.loads(response.text))
        response.close()
        return streamViewMap

    def getStreamViews(self, namespace_id, skip=0, count=100):
        """
        Retrieves a list of streamViews associated with the specified
            'namespace_id' under the current tenant
        :param namespace_id: namespace to work against
        :param skip: number of streamviews to skip for paging
        :param count: number streamviews in a page
        :return: array of SdsStreamviews
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__getStreamViewsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id),
            params={"skip": skip, "count": count})
        self.__baseClient.checkResponse(
            response, "Failed to get all SdsStreamViews.")

        content = json.loads(response.content)
        results = []
        for item in content:
            results.append(SdsStreamView.fromJson(item))
        response.close()
        return results

    def getOrCreateStreamView(self, namespace_id, streamView):
        """
        Tells Sds Service to create a streamView based on a local
            SdsStreamView object
        :param namespace_id: namespace to work against
        :param streamView: Streamview object to create in OCS
        :return: created Streamview as SdsStreamview
        """
        if namespace_id is None:
            raise TypeError
        if streamView is None or not isinstance(streamView, SdsStreamView):
            raise TypeError

        response = self.__baseClient.request(
            "post",
            self.__streamViewsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                streamView_id=streamView.Id),
            data=streamView.toJson())

        self.__baseClient.checkResponse(
            response, f"Failed to create SdsStreamView, {streamView.Id}.")

        streamView = SdsStreamView.fromJson(json.loads(response.text))
        response.close()
        return streamView

    def createOrUpdateStreamView(self, namespace_id, streamView):
        """
        Tells Sds Service to create a streamView based on a local
            SdsStreamView object
        :param namespace_id: namespace to work against
        :param streamView: Streamview object to create or update in OCS
        :return: created Streamview as SdsStream
        """
        if namespace_id is None:
            raise TypeError
        if streamView is None or not isinstance(streamView, SdsStreamView):
            raise TypeError

        response = self.__baseClient.request(
            "put",
            self.__streamViewsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                streamView_id=streamView.Id),
            data=streamView.toJson())
        self.__baseClient.checkResponse(
            response, f"Failed to create SdsStreamView, {streamView.Id}.")

        response.close()

    def deleteStreamView(self, namespace_id, streamView_id):
        """
        Tells Sds Service to delete the streamView with the specified
            'streamView_id'
        :param namespace_id: namespace to work against
        :param streamView_id: id of streamview to delete
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if streamView_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "delete",
            self.__streamViewsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                streamView_id=streamView_id))

        self.__baseClient.checkResponse(
            response, f"Failed to delete SdsStreamView, {streamView_id}.")

        response.close()

    def getStream(self, namespace_id, stream_id):
        """
        Retrieves a stream specified by 'stream_id' from the Sds Service
        :param namespace_id: namespace to work against
        :param stream_id: id of the stream
        :return:the Stream as SdsStream
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id))

        self.__baseClient.checkResponse(
            response, f"Failed to get SdsStream, {stream_id}.")

        stream = SdsStream.fromJson(json.loads(response.content))
        response.close()
        return stream

    def getStreamType(self, namespace_id, stream_id):
        """
        Retrieves a stream specified by 'stream_id' from the Sds Service
        :param namespace_id: namespace to work against
        :param stream_id: id of the stream
        :return: the stream type as an SdsType
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id) + "/Type")

        self.__baseClient.checkResponse(
            response, f"Failed to get SdsStream type,  {stream_id}.")

        type = SdsType.fromJson(json.loads(response.content))
        response.close()
        return type

    def getStreams(self, namespace_id, query="", skip=0, count=100):
        """
        Retrieves a list of streams associated with 'namespace_id' under
            the current tenant"
        :param namespace_id: namespace to work against
        :param query: filtering query
        :param skip: number of streams to skip for paging
        :param count: number of streams to limit to
        :return: array of SdsStreams
        """
        if namespace_id is None:
            raise TypeError
        if query is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__getStreamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id),
            params={"query": query, "skip": skip, "count": count})

        self.__baseClient.checkResponse(
            response, "Failed to get all SdsStreams.")

        content = json.loads(response.content)
        results: [SdsStream] = []
        for item in content:
            results.append(SdsStream.fromJson(item))
        response.close()
        return results

    def getOrCreateStream(self, namespace_id, stream):
        """
        Tells Sds Service to create a stream based on the local 'stream'
            SdsStream object
        :param namespace_id: namespace to work against
        :param stream: the stream to Create or retrieve, as a SDsStream
        :return: the created Stream as an SdsStream
        """
        if namespace_id is None:
            raise TypeError
        if stream is None or not isinstance(stream, SdsStream):
            raise TypeError
        response = self.__baseClient.request(
            "post",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream.Id),
            data=stream.toJson())

        self.__baseClient.checkResponse(
            response, f"Failed to create SdsStream, {stream.Id}.")

        stream = SdsStream.fromJson(json.loads(response.content))
        response.close()
        return stream

    def createOrUpdateStream(self, namespace_id, stream):
        """
        Tells Sds Service to create a stream based on the local 'stream'
            SdsStream object
        :param namespace_id: namespace to work against
        :param stream: the stream to Create or update, as a SDsStream
        :return: the created or updated Stream as an SdsStream
        """
        if namespace_id is None:
            raise TypeError
        if stream is None or not isinstance(stream, SdsStream):
            raise TypeError

        response = self.__baseClient.request(
            "put",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream.Id),
            data=stream.toJson())

        self.__baseClient.checkResponse(
            response, f"Failed to create SdsStream, {stream.Id}.")

        response.close()

    def updateStreamType(self, namespace_id, streamId, streamViewId):
        """
        Tells Sds Service to update a stream based on the local 'stream'
             SdsStream object
        :param namespace_id: namespace to work against
        :param stream_id: id of the stream to change the type of
        :param streamViewId: if of the streamview to change the type to
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if streamId is None:
            raise TypeError
        if streamViewId is None:
            raise TypeError

        response = self.__baseClient.request(
            "put",
            self.__updateStreamTypePath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=streamId),
            params={"streamViewId": streamViewId})

        self.__baseClient.checkResponse(
            response, f"Failed to update SdsStream type, {streamId}.")

        response.close()

    def deleteStream(self, namespace_id, stream_id):
        """
        Tells Sds Service to delete the stream speficied by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to delete
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "delete",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id))

        self.__baseClient.checkResponse(
            response, f"Failed to delete SdsStream, {stream_id}.")

        response.close()

    def createOrUpdateTags(self, namespace_id, streamId, tags):
        """
        "Tells Sds Service to create tags and associate them with the given
            streamId
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to update with tags
        :param tags: tags to create or update.  expected for is an array of
            strings
        :return:
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "put",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=streamId) + "/Tags",
            data=json.dumps(tags))

        self.__baseClient.checkResponse(
            response, f"Failed to create tags for Stream: {streamId}.")

    def createOrUpdateMetadata(self, namespace_id, streamId, metadata):
        """
        Tells Sds Service to create metadata and associate them with the given
            streamId
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to update with metadata
        :param metadata: metadata to create or update.  expected for is an
            dict(string,string)
        :return:
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "put",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=streamId) + "/Metadata",
            data=json.dumps(metadata))

        self.__baseClient.checkResponse(
            response, f"Failed to create metadata for Stream: {streamId}.")

    def patchMetadata(self, namespace_id, streamId, patch):
        """
        Tells Sds Service to update metadata on the given streamId
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to update with metadata
        :param patch: a JSON patch document
        :return:
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "patch",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=streamId) + "/Metadata",
            data=json.dumps(patch))

        self.__baseClient.checkResponse(
            response, f"Failed to update metadata for Stream: {streamId}.")

    def getTags(self, namespace_id, streamId):
        """
        Tells Sds Service to get tags associated with the given streamId
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the tags of
        :return: stream's tags
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=streamId) + "/Tags")

        self.__baseClient.checkResponse(
            response, f"Failed to get tags for Stream: {streamId}.")

        content = json.loads(response.text)
        response.close()
        return content

    def getMetadata(self, namespace_id, streamId, key):
        """
        Tells Sds Service to get metadata associated with the given streamId
            and key
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the metadata value of
        :param key: specific metadata field to retrieve
        :return: value at the key
        """
        if namespace_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__streamsPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=streamId) + "/Metadata/" + key)

        self.__baseClient.checkResponse(
            response, f"Failed to get metadata for Stream: {streamId}.")

        content = json.loads(response.text)
        response.close()
        return content

    # The following section provides functionality to interact with Data
    #  We assume the value(s) passed follow the Sds object patterns
    #  supporting fromJson and toJson method

    def getValue(self, namespace_id, stream_id, index, value_class=None):
        """
        Retrieves JSON object from Sds Service for value specified by 'index'
            from Sds Service
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the data of
        :param index: index at which to get a value
        :param value_class: use this to cast the value into a given type.
            Type must support .fromJson()  Default is None.
            If None returns a dynamic Python object from the data.
        :return: the value.  If value_class is defined it is in this type.
            Otherwise it is a dynamic Python object
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if index is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__dataPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            params={"index": index})

        self.__baseClient.checkResponse(
            response, f"Failed to get value for SdsStream: {stream_id}.")

        content = json.loads(response.content)
        response.close()

        if value_class is None:
            return content
        return value_class.fromJson(content)

    def getFirstValue(self, namespace_id, stream_id, value_class=None):
        """
        Retrieves JSON object from Sds Service the first value to be added to
            the stream specified by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the data of
        :param value_class: use this to cast the value into a given type.
            Type must support .fromJson()  Default is None.
            If None returns a dynamic Python object from the data.
        :return: the value.  If value_class is defined it is in this type.
            Otherwise it is a dynamic Python object
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__getFirstValue.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id))

        self.__baseClient.checkResponse(
            response, f"Failed to get first value for SdsStream: {stream_id}.")

        content = json.loads(response.content)
        response.close()
        if value_class is None:
            return content
        return value_class.fromJson(content)

    def getLastValue(self, namespace_id, stream_id, value_class=None):
        """
        Retrieves JSON object from Sds Service the last value to be added to
            the stream specified by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the data of
        :param value_class: use this to cast the value into a given type.
            Type must support .fromJson()  Default is None.
            If None returns a dynamic Python object from the data.
        :return: the value.  If value_class is defined it is in this type.
            Otherwise it is a dynamic Python object
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__getLastValue.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id))

        self.__baseClient.checkResponse(
            response, f"Failed to get last value for SdsStream: {stream_id}.")

        content = json.loads(response.text)
        response.close()
        if value_class is None:
            return content
        return value_class.fromJson(content)

    def getWindowValues(self, namespace_id, stream_id, value_class, start, end,
                        filter=""):
        """
        Retrieves JSON object representing a window of values from the stream
            specified by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the data of
        :param value_class: use this to cast the value into a given type.
            Type must support .fromJson().
            If None returns a dynamic Python object from the data.
        :param start: Starting index
        :param end: Ending index
        :param filter: An optional filter.  By Default it is "".
        :return: an array of values.
            If value_class is defined it is in this type.
            Otherwise it is a dynamic Python object
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if start is None:
            raise TypeError
        if end is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__dataPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            params={"startIndex": start, "endIndex": end, "filter": filter})

        self.__baseClient.checkResponse(
            response, f"Failed to get window values"
                      " for SdsStream: {stream_id}.")

        content = json.loads(response.text)
        response.close()
        if value_class is None:
            return content

        values = []
        for c in content:
            values.append(value_class.fromDictionary(c))
        return values

    def getWindowValuesForm(self, namespace_id, stream_id, value_class, start,
                            end, form=""):
        """
        Retrieves JSON object representing a window of values from the stream
            specified by 'stream_id'.  Use this to get the data in a different
            return form
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the data of
        :param value_class: use this to cast the value into a given type.
        Type must support .fromJson().
        If None returns a dynamic Python object from the data.
        :param start: Starting index
        :param end: Ending index
        :param form: form of the data
        :return: An array of the data in type specified if value_class
            defined.  Otherwise it is a dynamic Python object
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if start is None:
            raise TypeError
        if end is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__dataPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            params={"startIndex": start, "endIndex": end, "form": form})

        self.__baseClient.checkResponse(
            response, f"Failed to get window values for SdsStream, "
                      "form : {stream_id}.")

        content = json.loads(response.text)
        response.close()
        if value_class is None:
            return content

        values = []
        for c in content:
            values.append(value_class.fromDictionary(c))
        return values

    def getRangeValues(self, namespace_id, stream_id, value_class, start, skip,
                       count, reversed, boundary_type, streamView_id=""):
        """
        Retrieves JSON object representing a range of values from the stream
            specified by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the data of
        :param value_class: use this to cast the value into a given type.
            Type must support .fromJson(). If None returns a dynamic Python
            object from the data.
        :param start: Starting index
        :param skip: number of values to skip after start index.
            Important in paging
        :param count: number of values to return
        :param reversed: which direction to go when getting values
        :param boundary_type: the boundary condition to use.
            Can be an SdsBoundaryType or the integer value
        :param streamView_id: streamview to map the results to
        :return: An array of the data in type specified if value_class
            is defined.  Otherwise it is a dynamic Python object
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if start is None:
            raise TypeError
        if skip is None:
            raise TypeError
        if count is None:
            raise TypeError
        if reversed is None or not isinstance(reversed, bool):
            raise TypeError
        if boundary_type is None:
            raise TypeError

        boundary = boundary_type
        if isinstance(boundary_type, SdsBoundaryType):
            boundary = boundary_type.value

        response = self.__baseClient.request(
            "get",
            self.__transform.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            params={"startIndex": start, "skip": skip, "count": count,
                    "reversed": reversed, "boundary_type": boundary,
                    "streamView_id": streamView_id})

        self.__baseClient.checkResponse(
            response, f"Failed to get range values for"
            " SdsStream: {stream_id}.")

        content = json.loads(response.text)
        response.close()
        if value_class is None:
            return content
        values = []
        for c in content:
            values.append(value_class.fromJson(c))
        return values

    def getRangeValuesInterpolated(self, namespace_id, stream_id, value_class,
                                   start, end, count):
        """
        Retrieves JSON object representing a range of values from the stream
            specified by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the data of
        :param value_class: use this to cast the value into a given type.
            Type must support .fromJson(). If None returns a dynamic Python
            object from the data.
        :param start: starting index
        :param end:  ending index
        :param count: number of datapoints to retrieve
        :return: An array of the data in type specified if value_class is
        defined.  Otherwise it is a dynamic Python object
        """

        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if start is None:
            raise TypeError
        if end is None:
            raise TypeError
        if count is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__getRangeInterpolatedQuery.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            params={"startIndex": start, "endIndex": end, "count": count})

        self.__baseClient.checkResponse(
            response, f"Failed to get range values for"
                      " SdsStream: {stream_id}.")

        content = json.loads(response.text)
        response.close()
        if value_class is None:
            return content
        values = []
        for c in content:
            values.append(value_class.fromJson(c))
        return values

    def getSampledValues(self, namespace_id, stream_id, value_class, start,
                         end, sample_by, intervals, filter="",
                         stream_view_id=""):
        """
        Returns data sampled by intervals between a specified start and end
            index.
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the data of
        :param value_class: use this to cast the value into a given type.
            Type must support .fromJson().
            If None returns a dynamic Python object from the data.
        :param start: starting index for intervals
        :param end:  ending index for intervals
        :param sample_by: property or properties to use when sampling
        :param intervals: number of intervals requested
        :param boundary: optional SdsBoundaryType specifies the handling of
            events at or near the startIndex and endIndex
        :param start_boundary: optional SdsBoundaryType specifies the handling
            of events at or near the startIndex
        :param end_boundary: optional SdsBoundaryType specifies the handling
            of events at or near the endIndex
        :param filter: optional filter to apply
        :param stream_view_id: optional streamview identifier
        :return: An array of the data in type specified if value_class is
            defined.  Otherwise it is a dynamic Python object
        """

        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if start is None:
            raise TypeError
        if end is None:
            raise TypeError
        if sample_by is None:
            raise TypeError
        if intervals is None:
            raise TypeError

        # if stream_view_id is not set, do not specify /transform/ route
        # and stream_view_id parameter
        if len(stream_view_id) == 0:
            _path = self.__getSampledValues.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id)

        else:
            _path = self.__getSampledValuesT.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id)

        response = self.__baseClient.request(
            "get",
            _path,
            params={"startIndex": start,
                    "endIndex": end,
                    "sampleBy": sample_by,
                    "intervals": intervals,
                    "filter": filter,
                    "stream_view_id": stream_view_id})

        self.__baseClient.checkResponse(
            response, f"Failed to get sampled values for"
                      " SdsStream: {stream_id}.")

        content = json.loads(response.text)
        response.close()

        if value_class is None:
            return content
        values = []
        for c in content:
            values.append(value_class.fromJson(c))
        return values

    def getSummaries(self, namespace_id, stream_id, value_class, start, end,
                     count, stream_view_id="", filter=""):
        """
        Retrieves JSON object representing a summary for the stream specified
            by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get the data of
        :param value_class: use this to cast the value into a given type.
            Type must support .fromJson().
            If None returns a dynamic Python object from the data.
        Note- for this function the default return json string is not a
            JSON array of the value, so the same type definition won't work.
        :param start: starting index
        :param end:  ending index
        :param count: number of datapoints in summary
        :param stream_view_id: streamview to tranform the data into
        :param filter: filter to apply
        :return: An array of the data summary in type specified if value_class
            is defined.  Otherwise it is a dynamic Python object
        """

        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if start is None:
            raise TypeError
        if end is None:
            raise TypeError
        if count is None:
            raise TypeError

        # if stream_view_id is not set, do not specify /transform/ route
        # and stream_view_id parameter
        paramsToUse = {}
        if len(stream_view_id) == 0:
            _path = self.__getSummaries.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id)

            paramsToUse = {"startIndex": start,
                           "endIndex": end,
                           "count": count,
                           "filter": filter}
        else:
            _path = self.__getSummariesT.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id)

            paramsToUse = {"startIndex": start,
                           "endIndex": end,
                           "count": count,
                           "filter": filter,
                           "streamViewId": stream_view_id}

        response = self.__baseClient.request(
            "get",
            _path, paramsToUse)

        self.__baseClient.checkResponse(
            response, f"Failed to get summaries for SdsStream: {stream_id}.")

        content = json.loads(response.text)
        response.close()
        if value_class is None:
            return content

        values = []
        for c in content:
            values.append(value_class.fromDictionary(c))
        return values

    def insertValues(self, namespace_id, stream_id, values):
        """
        Tells Sds Service to insert the values, defined by the list 'values',
        into the stream specified by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get data into
        :param values: values to sends into OCS.
            Can be string of json array of values.
            Can be an array of values of a type that has toJson defined.
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if values is None:
            raise TypeError

        if callable(getattr(values[0], "toJson", None)):
            events = []
            for value in values:
                events.append(value.toDictionary())
            payload = json.dumps(events)
        else:
            payload = values

        response = self.__baseClient.request(
            "post",
            self.__insertValuesPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            data=payload)

        self.__baseClient.checkResponse(
            response, f"Failed to insert multiple values for "
                      " SdsStream: {stream_id}.")

    def updateValues(self, namespace_id, stream_id, values):
        """
        Tells Sds Service to update values defined by the SdsValue list,
            'values'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get data updated
        :param values: values to update in OCS.
        Can be string of json array of values.
        Can be an array of values of a type that has toJson defined.
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if values is None:
            raise TypeError

        if callable(getattr(values[0], "toJson", None)):
            events = []
            for value in values:
                events.append(value.toDictionary())
            payload = json.dumps(events)
        else:
            payload = values

        response = self.__baseClient.request(
            "put",
            self.__updateValuesPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            data=payload)

        self.__baseClient.checkResponse(
            response,  f"Failed to update all values "
                       " for SdsStream: {stream_id}.")

        response.close()

    def replaceValues(self, namespace_id, stream_id, values):
        """
        Tells Sds Service to replace the values defined by the list 'values'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get data replaced
        :param values: values to replace in OCS.
            Can be string of json array of values.
            Can be an array of values of a type that has toJson defined.
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if values is None:
            raise TypeError

        if callable(getattr(values[0], "toJson", None)):
            events = []
            for value in values:
                events.append(value.toDictionary())
            payload = json.dumps(events)
        else:
            payload = values

        response = self.__baseClient.request(
            "put",
            self.__replaceValuesPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            data=payload)

        self.__baseClient.checkResponse(
            response, f"Failed to replace values for SdsStream: {stream_id}.")

        response.close()

    def removeValue(self, namespace_id, stream_id, key):
        """
        Tells Sds Service to delete the value with a key property matching
            'key'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get data removed
        :param key: the index to remove
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if key is None:
            raise TypeError

        response = self.__baseClient.request(
            "delete",
            self.__dataPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            params={"index": key})

        self.__baseClient.checkResponse(
            response, f"Failed to remove value for SdsStream: {stream_id}.")

        response.close()

    def removeWindowValues(self, namespace_id, stream_id, start, end):
        """
        Tells Sds Service to delete a window of values in the stream specified
            by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_id: id of the stream to get data removed
        :param start: starting index
        :param end: ending index
        :return:
        """
        if namespace_id is None:
            raise TypeError
        if stream_id is None:
            raise TypeError
        if start is None:
            raise TypeError
        if end is None:
            raise TypeError

        response = self.__baseClient.request(
            "delete",
            self.__dataPath.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id,
                stream_id=stream_id),
            params={"startIndex": start, "endIndex": end})

        self.__baseClient.checkResponse(
            response, f"Failed to remove all values for "
                      " SdsStream: {stream_id}.")

        response.close()

    def getStreamsWindow(self, namespace_id, stream_ids, value_class, start,
                         end, joinMode=1):
        """
        Retrieves JSON object representing a window of values from the stream
             specified by 'stream_id'
        :param namespace_id: id of namespace to work against
        :param stream_ids: ids of the streams to get the data of
        :param value_class: use this to cast the value into a given type.
            Type must support .fromJson(). If None returns a dynamic Python
            object from the data.
        :param start: Starting index
        :param end: Ending index
        :param joinMode: Join mode, supports numbers or strings.
            Defaults to outer
        :return: an array of values.
            If value_class is defined it is in this type.
            Otherwise it is a dynamic Python object
        """
        if namespace_id is None:
            raise TypeError
        if stream_ids is None:
            raise TypeError
        if not stream_ids:
            raise TypeError
        if start is None:
            raise TypeError
        if end is None:
            raise TypeError
        if joinMode is None:
            raise TypeError

        response = self.__baseClient.request(
            "get",
            self.__bulkStreams.format(
                tenant_id=self.__tenant,
                namespace_id=namespace_id),
            params={"streams": ','.join(stream_ids),
                    "startIndex": start,
                    "endIndex": end,
                    "joinMode": joinMode})

        self.__baseClient.checkResponse(
            response, f"Failed to get bulk values for"
                      " SdsStream: {stream_ids}.")

        content = json.loads(response.text)
        response.close()
        if value_class is None:
            return content

        values = []
        for valueArray in content:
            valuesInside = []
            for value in valueArray:
                valuesInside.append(value_class.fromDictionary(value))
            values.append(valuesInside)
        return values

    # private methods

    def __setPathAndQueryTemplates(self):
        """
        creates the urls that are used
        :return:
        """
        self.__basePath = self.__uri_API + \
            "/Tenants/{tenant_id}/Namespaces/{namespace_id}"
        self.__getStreamViewsPath = self.__basePath + "/StreamViews"
        self.__streamViewsPath = self.__getStreamViewsPath + "/{streamView_id}"
        self.__getStreamsPath = self.__basePath + "/Streams"
        self.__streamsPath = self.__getStreamsPath + "/{stream_id}"
        self.__updateStreamTypePath = self.__streamsPath + "/Type"

        self.__dataPath = self.__streamsPath + "/Data"
        self.__getFirstValue = self.__dataPath + "/First"
        self.__getLastValue = self.__dataPath + "/Last"
        self.__transform = self.__dataPath + "/Transform"
        self.__getSummaries = self.__dataPath + "/Summaries"
        self.__getSummariesT = self.__transform + "/Summaries"
        self.__getSampledValues = self.__dataPath + "/Sampled"
        self.__getSampledValuesT = self.__transform + "/Sampled"
        self.__getRangeInterpolatedQuery = self.__transform + "/Interpolated"

        self.__insertValuesPath = self.__dataPath
        self.__updateValuesPath = self.__dataPath
        self.__replaceValuesPath = self.__dataPath + "?allowCreate=false"

        self.__bulkStreams = self.__basePath + "/Bulk/Streams/Data/Joins"
